@@ -13,7 +13,6 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
-#include <sys/wait.h>
 #include <limits.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -27,8 +26,7 @@
 #include "fs.h"
 #include "policy.h"
 #include "utils.h"
-
-#define MOUNTPROG "/bin/mount"
+#include "luks.h"
 
 /* error codes */
 const int E_ARGS = 1;
@@ -65,7 +63,9 @@ usage( const char* exename )
     "  Remove the lock on <device> for process <pid> again.\n\n"),
         exename);
     puts( _("Options:\n"
-    "  -a, --async : mount <device> with the 'async' option (default: 'sync')\n"
+    "  -r          : force <device> to be mounted read-only\n"
+    "  -w          : force <device> to be mounted read-write\n"
+    "  -s, --sync  : mount <device> with the 'sync' option (default: 'async')\n"
     "  --noatime   : mount <device> with the 'noatime' option (default: 'atime')\n"
     "  -e, --exec  : mount <device> with the 'exec' option (default: 'noexec')\n"
     "  -t <fs>     : mount as file system type <fs> (default: autodetected)\n"
@@ -73,6 +73,9 @@ usage( const char* exename )
     "                in an UTF-8 locale, otherwise mount default)\n"
     "  -u <umask>  : use specified umask instead of the default (only for\n"
     "                file sytems which actually support umask setting)\n"
+    " --passphrase <file>\n"
+    "                read passphrase from file instead of the terminal\n"
+    "                (only for LUKS encrypted devices)\n"
     "  -d, --debug : enable debug output (very verbose)\n"
     "  -h, --help  : print help message and exit successfuly") );
 }
@@ -195,6 +198,7 @@ do_mount_fstab( const char* device )
  *        caching)
  * @param noatime if not 0, the device will be mounted with 'noatime'
  * @param exec if not 0, the device will be mounted with 'exec'
+ * @param force_write 1 for forced r/w, 0 for forced r/o, -1 for kernel default
  * @param iocharset charset to use for file name conversion; NULL for mount
  *        default
  * @param umask User specified umask (NULL for default)
@@ -203,11 +207,9 @@ do_mount_fstab( const char* device )
  */
 int
 do_mount( const char* device, const char* mntpt, const char* fsname, int async,
-        int noatime, int exec, const char* iocharset, const char* umask, 
-        int suppress_errors )
+        int noatime, int exec, int force_write, const char* iocharset, const
+        char* umask, int suppress_errors )
 {
-    int status;
-    int devnull;
     const struct FS* fs;
     char ugid_opt[100];
     char umask_opt[100];
@@ -215,6 +217,7 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
     const char* sync_opt = ",sync";
     const char* atime_opt = ",atime";
     const char* exec_opt = ",noexec";
+    const char* access_opt = NULL;
     char options[1000];
 
     /* check and retrieve option information for requested file system */
@@ -254,6 +257,13 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
     if( exec )
         exec_opt = ",exec";
 
+    if( force_write == 0 )
+        access_opt = ",ro";
+    else if( force_write == 1 )
+        access_opt = ",rw";
+    else
+        access_opt = "";
+
     if( iocharset && fs->support_iocharset ) {
         if( !is_word_str( iocharset ) ) {
             fprintf( stderr, _("Error: invalid charset name '%s'\n"), iocharset );
@@ -262,43 +272,14 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
         snprintf( iocharset_opt, sizeof( iocharset_opt ), ",iocharset=%s", iocharset );
     }
 
-    snprintf( options, sizeof( options ), "%s%s%s%s%s%s%s", 
-            fs->options, sync_opt, atime_opt, exec_opt, ugid_opt, umask_opt, iocharset_opt );
+    snprintf( options, sizeof( options ), "%s%s%s%s%s%s%s%s", 
+            fs->options, sync_opt, atime_opt, exec_opt, access_opt, ugid_opt,
+            umask_opt, iocharset_opt );
 
     /* go for it */
-    debug( "attempting mount: executing '%s %s %s %s %s %s %s'\n", MOUNTPROG,
-                "-t", fsname, "-o", options, device, mntpt );
-
-    if( !fork() ) {
-        get_root();
-        if( setreuid( 0, 0 ) ) {
-            perror( _("Error: could not raise to full root uid privileges") );
-            exit( 100 );
-        }
-
-        if( suppress_errors ) {
-            /* try to reroute stderr to /dev/null to suppress error messages */
-            devnull = open( "/dev/null", O_WRONLY );
-            if( devnull > 0 )
-                dup2( devnull, 2 );
-        }
-
-        execl( MOUNTPROG, MOUNTPROG, "-t", fsname, "-o", options, device, mntpt, NULL );
-        perror( _("Error: could not execute mount") );
-        exit( E_EXECMOUNT );
-    } else {
-        if( wait( &status ) < 0 ) {
-            perror( _("Error: could not wait for executed mount process") );
-            exit( E_EXECMOUNT );
-        }
-    }
-
-    debug( "mount attempt terminated with status %i\n", status );
-
-    if( WIFEXITED( status ) )
-        return WEXITSTATUS( status );
-    else
-        return -1;
+    return spawn( SPAWN_EROOT | SPAWN_RROOT | (suppress_errors ? SPAWN_NO_STDERR : 0 ),
+             MOUNTPROG, MOUNTPROG, "-t", fsname, "-o", options, device, mntpt,
+             NULL );
 }
 
 /**
@@ -310,6 +291,7 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
  *        caching)
  * @param noatime if not 0, the device will be mounted with 'noatime'
  * @param exec if not 0, the device will be mounted with 'exec'
+ * @param force_write 1 for forced r/w, 0 for forced r/o, -1 for kernel default
  * @param iocharset charset to use for file name conversion; NULL for mount
  *        default
  * @param umask User specified umask (NULL for default)
@@ -317,7 +299,8 @@ do_mount( const char* device, const char* mntpt, const char* fsname, int async,
  */
 int
 do_mount_auto( const char* device, const char* mntpt, int async, 
-        int noatime, int exec, const char* iocharset, const char* umask )
+        int noatime, int exec, int force_write, const char* iocharset, 
+        const char* umask )
 {
     const struct FS* fs;
     int nostderr = 1;
@@ -328,14 +311,14 @@ do_mount_auto( const char* device, const char* mntpt, int async,
         if( (fs+1)->fsname == NULL )
             nostderr = 0;
         result = do_mount( device, mntpt, fs->fsname, async, noatime, exec,
-                iocharset, umask, nostderr );
+                force_write, iocharset, umask, nostderr );
         if( result == 0 )
             break;
 
 	/* sometimes VFAT fails when using iocharset; try again without */
 	if( iocharset )
 	    result = do_mount( device, mntpt, fs->fsname, async, noatime, exec,
-                    NULL, umask, nostderr );
+                    force_write, NULL, umask, nostderr );
         if( result == 0 )
             break;
     }
@@ -494,14 +477,17 @@ main( int argc, char** argv )
     char *devarg = NULL, *arg2 = NULL;
     char mntpt[MEDIA_STRING_SIZE];
     char device[PATH_MAX], mntptdev[PATH_MAX];
+    char decrypted_device[PATH_MAX];
     const char* fstab_device;
     int is_real_path = 0;
-    int async = 0;
+    int async = 1;
     int noatime = 0;
     int exec = 0;
+    int force_write = -1; /* 0: ro, 1: rw, -1: default */
     const char* use_fstype = NULL;
     const char* iocharset = NULL;
     const char* umask = NULL;
+    const char* passphrase = NULL;
     int result;
 
     enum { MOUNT, LOCK, UNLOCK } mode = MOUNT;
@@ -512,12 +498,15 @@ main( int argc, char** argv )
         { "debug", 0, NULL, 'd'},
         { "lock", 0, NULL, 'l'},
         { "unlock", 0, NULL, 'L'},
-        { "async", 0, NULL, 'a' },
+        { "sync", 0, NULL, 's' },
         { "noatime", 0, NULL, 'A' },
         { "exec", 0, NULL, 'e' },
         { "type", 1, NULL, 't' },
         { "charset", 1, NULL, 'c' },
         { "umask", 1, NULL, 'u' },
+        { "passphrase", 1, NULL, 'p' },
+        { "read-only", 0, NULL, 'r' },
+        { "read-write", 0, NULL, 'w' },
         { NULL, 0, NULL, 0}
     };
 
@@ -537,7 +526,7 @@ main( int argc, char** argv )
 
     /* parse command line options */
     do {
-        switch( option = getopt_long( argc, argv, "+hdelLaAt:c:u:", long_opts, NULL ) ) {
+        switch( option = getopt_long( argc, argv, "+hdelLsArwt:c:u:", long_opts, NULL ) ) {
             case -1:  break;          /* end of arguments */
             case ':':
             case '?': return E_ARGS;  /* unknown argument */
@@ -550,7 +539,7 @@ main( int argc, char** argv )
 
             case 'L': mode = UNLOCK; break;
 
-            case 'a': async = 1; break;
+            case 's': async = 0; break;
 
             case 'A': noatime = 1; break;
 
@@ -562,6 +551,12 @@ main( int argc, char** argv )
 
             case 'u': umask = optarg; break;
 
+            case 'p': passphrase = optarg; break;
+
+            case 'r': force_write = 0; break;
+
+            case 'w': force_write = 1; break;
+
             default:
                 fprintf( stderr, _("Internal error: getopt_long() returned unknown value\n") );
                 return E_INTERNAL;
@@ -647,15 +642,38 @@ main( int argc, char** argv )
             if( check_mount_policy( device, mntpt )  )
                 return E_POLICY;
 
+            /* check for encrypted device */
+            enum decrypt_status decrypt = luks_decrypt( device,
+                    decrypted_device, sizeof( decrypted_device ), passphrase ); 
+
+            switch (decrypt) {
+                case DECRYPT_FAILED:
+                    fprintf( stderr, _("Error: could not decrypt device (wrong passphrase?)\n") );
+                    exit( E_POLICY );
+                case DECRYPT_EXISTS:
+                    fprintf( stderr, _("Error: mapped device already exists\n") );
+                    exit( E_POLICY );
+                case DECRYPT_OK:
+                case DECRYPT_NOTENCRYPTED:
+                    break;
+                default:
+                    fprintf( stderr, "Internal error: unhandled decrypt_status %i\n", 
+                        (int) decrypt);
+                    exit( E_INTERNAL );
+            }
+
             /* off we go */
             if( use_fstype )
-                result = do_mount( device, mntpt, use_fstype, async, noatime,
-                        exec, iocharset, umask, 0 );
+                result = do_mount( decrypted_device, mntpt, use_fstype, async, noatime,
+                        exec, force_write, iocharset, umask, 0 );
             else
-                result = do_mount_auto( device, mntpt, async, noatime, exec,
-                        iocharset, umask ); 
+                result = do_mount_auto( decrypted_device, mntpt, async, noatime, exec,
+                        force_write, iocharset, umask ); 
 
             if( result ) {
+                if( decrypt == DECRYPT_OK )
+                    luks_release( decrypted_device );
+
                 /* mount failed, delete the mount point again */
                 if( remove_pmount_mntpt( mntpt ) ) {
                     perror( _("Error: could not delete mount point") );
